(*
 * ocamlsdl - Objective Caml interface for the SDL library
 * Copyright (C) 1999 Jean-Christophe FILLIATRE
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License version 2, as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU Library General Public License version 2 for more details
 * (enclosed in the file LGPL).
 *)

(* Interface to the SDL mixer library (higher level than Audio) *)

exception Mix_failure of string

type format

val format_AUDIO_U8	: format
val format_AUDIO_S8	: format
val format_AUDIO_U16LSB	: format
val format_AUDIO_S16LSB	: format
val format_AUDIO_U16MSB	: format
val format_AUDIO_S16MSB	: format
val format_AUDIO_U16	: format
val format_AUDIO_S16	: format

val open_audio : int -> format -> int -> int -> unit
    (* [open_audio frequency format channels chunksize] *)

val close_audio : unit -> unit

type music

val load_music : string -> music

val play_music : music -> int -> int
    (* returns the channel on which it is played *)

val playing_music : unit -> bool

(* Halt a channel, fading it out progressively till it's silent
   The ms parameter indicates the number of milliseconds the fading
   will take.
 *)
(*
extern DECLSPEC int Mix_FadeOutChannel(int which, int ms);
extern DECLSPEC int Mix_FadeOutGroup(int tag, int ms);
*)

val fade_out_music : int -> unit

val halt_music : unit -> unit

val free_music : music -> unit


type chunk

val load_wav : string -> chunk

val play_channel : int -> chunk -> int -> int
    (* [play_channel channel chunk loops]. 
       returns the channel on which it is played *)

val playing : int -> bool
    (* check all if -1 *)

val halt_channel : int -> unit
    (* all if -1 *)

val free_chunk : chunk -> unit


