
(* lesson 02 *)

open Sdl
open Video
open Event

open GlDraw

let init_gl width height =
  (* We call this right after our OpenGL window is created. *)
  GlDraw.viewport 0 0 width height;
  (* This Will Clear The Background Color To Black *)
  GlClear.color ~alpha:0.0 (0.0, 0.0, 0.0);
  (* Enables Clearing Of The Depth Buffer *)
  GlClear.depth 1.0;
  (* The Type Of Depth Test To Do *)
  GlFunc.depth_func `less;
  (* Enables Depth Testing *)
  Gl.enable `depth_test;
  (* Enables Smooth Color Shading *)
  GlDraw.shade_model `smooth;
  GlMat.mode `projection;
  (* Reset The Projection Matrix *)
  GlMat.load_identity ();
  (* Calculate The Aspect Ratio Of The Window *)
  GluMat.perspective ~fovy:45.0 ~aspect:(float width /. float height) 
    ~z:(0.1, 100.0);
  GlMat.mode `modelview

(* The main drawing function. *)
let draw_gl_scene () =
   GlClear.clear [`color; `depth]; (* Clear The Screen And The Depth Buffer *)
  GlMat.load_identity ();				(* Reset The View *)

  GlMat.translate3 (-1.5, 0.0, -6.0);		(* Move Left 1.5 Units And Into The Screen 6.0 *)
	
  (* draw a triangle *)
  GlDraw.begins `polygon;				(* start drawing a polygon *)
  GlDraw.vertex3 ( 0.0, 1.0, 0.0);		(* Top *)
  GlDraw.vertex3 ( 1.0,-1.0, 0.0);		(* Bottom Right *)
  GlDraw.vertex3 (-1.0,-1.0, 0.0);		(* Bottom Left	*)
  GlDraw.ends ();					(* we're done with the polygon *)

  GlMat.translate3 (3.0, 0.0, 0.0);		        (* Move Right 3 Units *)
	
  (* draw a square (quadrilateral) *)
  GlDraw.begins `quads;				(* start drawing a polygon (4 sided) *)
  GlDraw.vertex3 (-1.0, 1.0, 0.0);		(* Top Left *)
  GlDraw.vertex3 ( 1.0, 1.0, 0.0);		(* Top Right *)
  GlDraw.vertex3 ( 1.0,-1.0, 0.0);		(* Bottom Right *)
  GlDraw.vertex3 (-1.0,-1.0, 0.0);		(* Bottom Left	*)
  GlDraw.ends ();				(* done with the polygon *)

  (* swap buffers to display, since we're double buffered. *)
  Sdl.GL.swap_buffers ()

let main () =
  let done_ = ref false in
  (* Initialize SDL for video output *)
  Sdl.init [VIDEO];
  (* Create a 640x480 OpenGL screen *)
  let screen = set_video_mode 640 480 0 [OPENGL] in
  (* Loop, drawing and checking events *)
  init_gl 640 480;
  while not !done_ do
    draw_gl_scene ();
    (* This could go in a separate function *)
    match poll_event () with
      | Some Quit -> done_ := true
      | Some (KeyDown k) ->
	  (match k.key_sym.key with
	     | 27 -> done_ := true
	     | _ -> ())
      | _ -> ()
  done

let _ = try main (); Sdl.quit () with e -> Sdl.quit (); raise e
