/*
 * ocamlsdl - Objective Caml interface for the SDL library
 * Copyright (C) 1999 Jean-Christophe FILLIATRE
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License version 2, as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * See the GNU Library General Public License version 2 for more details
 * (enclosed in the file LGPL).
 */

/* $Id: sdl_stub.c,v 1.12 2002/12/10 16:36:14 filliatr Exp $ */

#include <stdio.h>
#include <SDL.h>
#include <caml/mlvalues.h>
#include <caml/memory.h>
#include <caml/alloc.h>
#include <caml/fail.h>
#include <caml/callback.h>

/*

  Caml list manipulations

*/

#define NIL_tag 0
#define CONS_tag 1

value nil(void)
{
  return Val_int(0);
}

value cons(value x,value l)
{
  CAMLparam2(x,l);
  CAMLlocal1(m);
  m=alloc(2,CONS_tag);
  Store_field(m,0,x);
  Store_field(m,1,l);
  CAMLreturn (m);
}

int is_nil(value l)
{
  return Is_long(l);
}

int is_not_nil(value l)
{
  return Is_block(l);
}

value hd(value l)
{
  return Field(l,0);
}

value tl(value l)
{
  return Field(l,1);
}

/*

  conversion between OCAMLSDL flags and C SDL flags

*/
  
#define TIMER_tag 0
#define AUDIO_tag 1
#define VIDEO_tag 2
#define CDROM_tag 3
#define JOYSTICK_tag 4
#define NOPARACHUTE_tag 5       
#define EVENTTHREAD_tag 6
#define EVERYTHING_tag 7

int init_flag_val(value flag_list)
{
  int flag = 0;
  value l = flag_list;
  while (is_not_nil(l))
    {
      switch (Int_val(hd(l)))
	{
	case TIMER_tag       : flag |= SDL_INIT_TIMER       ; break;
	case AUDIO_tag       : flag |= SDL_INIT_AUDIO       ; break;
	case VIDEO_tag       : flag |= SDL_INIT_VIDEO       ; break;
	case CDROM_tag       : flag |= SDL_INIT_CDROM       ; break;
	case JOYSTICK_tag    : flag |= SDL_INIT_JOYSTICK    ; break;
	case NOPARACHUTE_tag : flag |= SDL_INIT_NOPARACHUTE ; break;
	case EVENTTHREAD_tag : flag |= SDL_INIT_EVENTTHREAD ; break;
	case EVERYTHING_tag  : flag |= SDL_INIT_EVERYTHING  ; break;
	}
      l = tl(l);
    }
  return flag;
}

#define SWSURFACE_tag 0 
#define HWSURFACE_tag 1   
#define ANYFORMAT_tag 2  
#define HWPALETTE_tag 3   
#define DOUBLEBUF_tag 4   
#define FULLSCREEN_tag 5  
#define HWACCEL_tag 6     
#define SRCCOLORKEY_tag 7  
#define RLEACCEL_tag 8    
#define SRCALPHA_tag 9    
#define SRCCLIPPING_tag 10  
#define OPENGL_tag 11

int video_flag_val(value flag_list)
{
  int flag = 0;
  value l = flag_list;
  while (is_not_nil(l))
    {
      switch (Int_val(hd(l)))
	{
	case SWSURFACE_tag   : flag |= SDL_SWSURFACE   ; break;
	case HWSURFACE_tag   : flag |= SDL_HWSURFACE   ; break;
	case ANYFORMAT_tag   : flag |= SDL_ANYFORMAT   ; break;
	case HWPALETTE_tag   : flag |= SDL_HWPALETTE   ; break;
	case DOUBLEBUF_tag   : flag |= SDL_DOUBLEBUF   ; break;
	case FULLSCREEN_tag  : flag |= SDL_FULLSCREEN  ; break;
	case HWACCEL_tag     : flag |= SDL_HWACCEL     ; break;
	case SRCCOLORKEY_tag : flag |= SDL_SRCCOLORKEY ; break;
	case RLEACCEL_tag    : flag |= SDL_RLEACCEL    ; break;
	case SRCALPHA_tag    : flag |= SDL_SRCALPHA    ; break;
	  /*	case SRCCLIPPING_tag : flag |= SDL_SRCCLIPPING ; break; */
	case OPENGL_tag      : flag |= SDL_OPENGL      ; break; 
	}
      l = tl(l);
    }
  return flag;
}

value val_video_flag(int flags)
{
  value l = nil();
  if (flags & SDL_SWSURFACE)   l = cons(Val_int(SWSURFACE_tag),l);
  if (flags & SDL_HWSURFACE)   l = cons(Val_int(HWSURFACE_tag),l);
  if (flags & SDL_ANYFORMAT)   l = cons(Val_int(ANYFORMAT_tag),l);
  if (flags & SDL_HWPALETTE)   l = cons(Val_int(HWPALETTE_tag),l);
  if (flags & SDL_DOUBLEBUF)   l = cons(Val_int(DOUBLEBUF_tag),l);
  if (flags & SDL_FULLSCREEN)  l = cons(Val_int(FULLSCREEN_tag),l);
  if (flags & SDL_HWACCEL)     l = cons(Val_int(HWACCEL_tag),l);
  if (flags & SDL_SRCCOLORKEY) l = cons(Val_int(SRCCOLORKEY_tag),l);
  if (flags & SDL_RLEACCEL)    l = cons(Val_int(RLEACCEL_tag),l);
  if (flags & SDL_SRCALPHA)    l = cons(Val_int(SRCALPHA_tag),l);
  /*  if (flags & SDL_SRCCLIPPING) l = cons(Val_int(SRCCLIPPING_tag),l); */
  if (flags & SDL_OPENGL)      l = cons(Val_int(OPENGL_tag),l);
  return l;
}

/*

  raising SDL_failure exception

*/

void raise_failure() {
  raise_with_string(*caml_named_value("SDL_failure"), SDL_GetError());
}

value sdlstub_init(value vf) {
  int flags = init_flag_val(vf);

  if (SDL_Init(flags) < 0) raise_failure();
  return Val_unit;
}

value sdlstub_quit(value u) {
  SDL_Quit();
  return Val_unit;
}

value sdlstub_must_lock(value s) {
  int b = SDL_MUSTLOCK(((SDL_Surface*) s));

  return Val_bool(b);
}

value sdlstub_video_mode_ok(value vw, value vh, value vbpp, value vf) {
  int w = Int_val(vw);
  int h = Int_val(vh);
  int bpp = Int_val(vbpp);
  int flags = video_flag_val(vf);

  return Val_bool(SDL_VideoModeOK(w,h,bpp,flags));
}

value sdlstub_set_video_mode(value vw, value vh, value vbpp, value vf) {
  int w = Int_val(vw);
  int h = Int_val(vh);
  int bpp = Int_val(vbpp);
  int flags = video_flag_val(vf);
  SDL_Surface* s;
  
  if ((s = SDL_SetVideoMode(w,h,bpp,flags)) == NULL) raise_failure();
  return (value) s;
}

value sdlstub_load_bmp(value vfile) {
  char * file = String_val(vfile);
  SDL_Surface* s;

  if ((s = SDL_LoadBMP(file)) == NULL) raise_failure();
  return (value) s;
}

value sdlstub_save_bmp(value s, value vfile) {
  char * file = String_val(vfile);

  if (SDL_SaveBMP((SDL_Surface*) s, file) < 0) raise_failure();
  return Val_unit;
}

/* Code by Jeff Molofee's openGL tutorial */
SDL_Surface * GLLoadBMP(char *filename)
{
    Uint8 *rowhi, *rowlo;
    Uint8 *tmpbuf, tmpch;
    SDL_Surface *image;
    int i, j;

    image = SDL_LoadBMP(filename);
    if (image == NULL) {
        fprintf(stderr, "Unable to load %s: %s\n", filename, SDL_GetError());
        raise_failure();
    }

    /* GL surfaces are upsidedown and RGB, not BGR :-) */
    tmpbuf = (Uint8 *)malloc(image->pitch);
    if (tmpbuf == NULL) {
        fprintf(stderr, "Out of memory\n");
        raise_failure();
    }
    rowhi = (Uint8 *)image->pixels;
    rowlo = rowhi + (image->h * image->pitch) - image->pitch;
    for ( i=0; i<image->h/2; ++i ) {
        for ( j=0; j<image->w; ++j ) {
            tmpch = rowhi[j*3];
            rowhi[j*3] = rowhi[j*3+2];
            rowhi[j*3+2] = tmpch;
            tmpch = rowlo[j*3];
            rowlo[j*3] = rowlo[j*3+2];
            rowlo[j*3+2] = tmpch;
        }
        memcpy(tmpbuf, rowhi, image->pitch);
        memcpy(rowhi, rowlo, image->pitch);
        memcpy(rowlo, tmpbuf, image->pitch);
        rowhi += image->pitch;
        rowlo -= image->pitch;
    }
    free(tmpbuf);
    return(image);
}

value sdlstub_GL_load_bmp(value f) {
  return (value)GLLoadBMP(String_val(f));
}

value sdlstub_string_of_pixels(value s) {
  CAMLparam1(s);
  CAMLlocal1(v);
  SDL_Surface * surf = (SDL_Surface *) s;
  int n = surf->w * surf->h * surf->format->BytesPerPixel;
  v = alloc_string(n);
  memcpy(String_val(v), surf->pixels, n);
  CAMLreturn(v);
}

value sdlstub_set_color_key(value s, value vf, value vk) {
  int flag = video_flag_val(vf);
  int key = Int_val(vk);

  if (SDL_SetColorKey((SDL_Surface*) s, flag, key) < 0) raise_failure();
  return Val_unit;
}

value sdlstub_set_alpha(value s, value vf, value va) {
  int flags = Int_val(vf);
  int alpha = Int_val(va);

  if (SDL_SetAlpha((SDL_Surface*) s, flags, alpha) < 0) raise_failure();
  return Val_unit;
}

/*
value sdlstub_set_clipping(value s, value vtop, value vleft, 
			   value vbottom, value vright) {
  int top = Int_val(vtop);
  int left = Int_val(vleft);
  int bottom = Int_val(vbottom);
  int right = Int_val(vright);
  
  SDL_SetClipping((SDL_Surface*) s, top, left, bottom, right);
  return Val_unit;
}
*/

value sdlstub_display_format(value s) {
  SDL_Surface* n;

  if ((n = SDL_DisplayFormat((SDL_Surface*) s)) == NULL) raise_failure();
  return (value) n;
}

value sdlstub_alloc_surface(value vflags, value vw, value vh, value vdepth, 
			    value vrmask, value vgmask, value vbmask,
			    value vamask) {
  SDL_Surface* s;
  Uint32 flags = Int_val(vflags);
  int width = Int_val(vw);
  int height = Int_val(vh);
  int depth = Int_val(vdepth);
  Uint32 rmask = Int_val(vrmask);
  Uint32 gmask = Int_val(vgmask);
  Uint32 bmask = Int_val(vbmask);
  Uint32 amask = Int_val(vamask);
  
  s = SDL_AllocSurface(flags, width, height, depth, 
		       rmask, gmask, bmask, amask); 
  if (s == NULL) raise_failure(); 
  return (value) s;
}

value sdlstub_alloc_surface_byte(value * argv, int n) {
  return sdlstub_alloc_surface(argv[0], argv[1], argv[2], argv[3], argv[4], 
			       argv[5], argv[6], argv[7]);
}

value sdlstub_free_surface(value s) {
  SDL_FreeSurface((SDL_Surface*) s);
  return Val_unit;
}

value sdlstub_lock_surface(value s) {
  if (SDL_LockSurface((SDL_Surface*) s) < 0) raise_failure();
  return Val_unit;
}

value sdlstub_unlock_surface(value s) {
  SDL_UnlockSurface((SDL_Surface*) s);
  return Val_unit;
}

value sdlstub_surface_width(value s) {
  return Val_int(((SDL_Surface*) s)->w);
}

value sdlstub_surface_height(value s) {
  return Val_int(((SDL_Surface*) s)->h);
}

value sdlstub_surface_flags(value s) {
  return val_video_flag(((SDL_Surface*) s)->flags);
}

value sdlstub_surface_bpp(value s) {
  return Val_int(((SDL_Surface*) s)->format->BitsPerPixel);
}

value sdlstub_surface_rmask(value s) {
  return Val_int(((SDL_Surface*) s)->format->Rmask);
}

value sdlstub_surface_gmask(value s) {
  return Val_int(((SDL_Surface*) s)->format->Gmask);
}

value sdlstub_surface_bmask(value s) {
  return Val_int(((SDL_Surface*) s)->format->Bmask);
}

value sdlstub_surface_amask(value s) {
  return Val_int(((SDL_Surface*) s)->format->Amask);
}

value sdlstub_map_rgb(value s, value vr, value vg, value vb) {
  int r = Int_val(vr);
  int g = Int_val(vg);
  int b = Int_val(vb);

  return Val_int(SDL_MapRGB(((SDL_Surface*) s)->format, r, g, b));
}

value sdlstub_get_ticks(value u) {
  return Val_int(SDL_GetTicks());
}

value sdlstub_delay(value vms) {
  int ms = Int_val(vms);

  SDL_Delay(ms);
  return Val_unit;
}

value sdlstub_fill_surface(value s, value vc) {
  int c = Int_val(vc);

  if (SDL_FillRect((SDL_Surface*) s, NULL, c) < 0) raise_failure();
  return Val_unit;
}

value sdlstub_fill_rect(value s, value vr, value vc) {
  int c = Int_val(vc);
  SDL_Rect r;
  
  r.x = Int_val(Field(vr,0));
  r.y = Int_val(Field(vr,1));
  r.w = Int_val(Field(vr,2));
  r.h = Int_val(Field(vr,3));
  if (SDL_FillRect((SDL_Surface*) s, &r, c) < 0) raise_failure();
  return Val_unit;
}

value sdlstub_update_surface(value s) {
  SDL_UpdateRect((SDL_Surface*) s, 0, 0, 0, 0);
  return Val_unit;
}

value sdlstub_update_rect(value s, value vx, value vy, value vw, value vh) {
  int x = Int_val(vx);
  int y = Int_val(vy);
  int w = Int_val(vw);
  int h = Int_val(vh);
  
  SDL_UpdateRect((SDL_Surface*) s, x, y, w, h);
  return Val_unit;
}

value sdlstub_update_rects(value s, value vn, value arr){
  int n = Int_val(vn);
  value v;
  int i;
  SDL_Rect* rects;

  if (Wosize_val(arr) < n) invalid_argument("update_rects");
  rects = (SDL_Rect*) malloc(n * sizeof(SDL_Rect));
  for (i = 0; i < n; i++) {
    v = Field(arr,i);
    rects[i].x = Int_val(Field(v,0));
    rects[i].y = Int_val(Field(v,1));
    rects[i].w = Int_val(Field(v,2));
    rects[i].h = Int_val(Field(v,3));
  };
  SDL_UpdateRects((SDL_Surface*) s, n, rects);
  free(rects);
  return Val_unit;
}

value sdlstub_flip(value s) {
  if (SDL_Flip((SDL_Surface*) s) < 0) raise_failure();
  return Val_unit;
}

SDL_Rect* rect_from_option(value v, SDL_Rect* r) {
  value vr;

  if (v == Val_int(0)) { 
    /* None */
    return NULL;
  } else {
    /* Some */
    vr = Field(v,0);
    r->x = Int_val(Field(vr,0));
    r->y = Int_val(Field(vr,1));
    r->w = Int_val(Field(vr,2));
    r->h = Int_val(Field(vr,3));
    return r;
  }
}

/* assumption: v = Some rect  and  r is not NULL */
void update_rect_option(value v, SDL_Rect* r) {
  value vr = Field(v,0);
  modify(&Field(vr,0), Val_int(r->x));
  modify(&Field(vr,1), Val_int(r->y));
  modify(&Field(vr,2), Val_int(r->w));
  modify(&Field(vr,3), Val_int(r->h));
}

value sdlstub_blit_surface(value src, value srcr, value dst, value dstr) {
  SDL_Rect sr,dr;
  SDL_Rect *srp, *drp;

  srp = rect_from_option(srcr,&sr);
  drp = rect_from_option(dstr,&dr);
  if (SDL_BlitSurface((SDL_Surface*) src, srp, (SDL_Surface*) dst, drp) < 0)
    raise_failure();
  if (! (srp == NULL)) update_rect_option(srcr,srp);
  if (! (drp == NULL)) update_rect_option(dstr,drp);
  return Val_unit;
}

value value_for_keysym(SDL_Event* event) {
  CAMLparam0();
  CAMLlocal1(v);
  SDL_keysym ks = event->key.keysym;
  v = alloc_tuple(4);
  Store_field(v,0,Val_int(ks.scancode));
  Store_field(v,1,Val_int(ks.sym));
  Store_field(v,2,Val_int(ks.mod));
  Store_field(v,3,Val_int(ks.unicode));
  CAMLreturn (v);
}

value value_for_keyboard(SDL_Event* event) {
  CAMLparam0();
  CAMLlocal2(v,ks);

  ks = value_for_keysym(event);
  v = alloc_tuple(2);
  Store_field(v,0,Val_int((event->key.state == SDL_PRESSED) ? 0 : 1));
  Store_field(v,1,ks);
  CAMLreturn (v);
}

/* returns (C_tag v) */
value unary_constructor(value v, int tag) {
  CAMLparam1(v);
  CAMLlocal1(result);
  result = alloc(1,tag);
  Store_field(result,0,v);
  CAMLreturn (result);
}

value value_for_event(SDL_Event* event) {
  int tag;
  value v;

  switch (event->type) {
  case SDL_ACTIVEEVENT:
    return Val_int(0);
  case SDL_KEYDOWN:
  case SDL_KEYUP:
    v = value_for_keyboard(event);
    tag = ((event->type == SDL_KEYDOWN) ? 0 : 1);
    return unary_constructor(v,tag);
  case SDL_QUIT:
    return Val_int(7);
  default:
    return Val_int(9);
  }
}

value sdlstub_poll_event(value u) {
  CAMLparam1(u);
  CAMLlocal2(v,result);
  SDL_Event event;

  if (SDL_PollEvent(&event)) {
    /* Some */
    v = value_for_event(&event);
    result = alloc(1,0);
    Store_field(result,0,v);
    CAMLreturn (result);
  } else { 
    /* None */
    CAMLreturn (Val_int(0));
  }
}

value sdlstub_wait_event(value u) {
  CAMLparam1(u);
  SDL_Event event;
  value v;

  if (SDL_WaitEvent(&event)) {
    CAMLreturn (value_for_event(&event));
  } else { 
    raise_failure();
    CAMLreturn (Val_unit);
  }
}

value sdlstub_set_colors(value s, value arr, value vfirst, value vn) {
  int ncolors = Int_val(vn);
  int firstcolor = Int_val(vfirst);
  int result, i;
  value v;
  SDL_Color* colors;

  if (ncolors < 0 || ncolors > Wosize_val(arr)) invalid_argument("set_colors");
  colors = (SDL_Color*) malloc(ncolors * sizeof(SDL_Color));
  if (colors == NULL) raise_out_of_memory();
  for (i=0; i<ncolors; i++) {
    v = Field(arr,i);
    colors[i].r = Int_val(Field(v,0));
    colors[i].g = Int_val(Field(v,1));
    colors[i].b = Int_val(Field(v,2));
  };
  result = SDL_SetColors((SDL_Surface*) s, colors, firstcolor, ncolors);
  free(colors);
  return Val_bool(result);
}

value sdlstub_show_cursor(value vtoggle) {
  int toggle = Bool_val(vtoggle);
  
  SDL_ShowCursor(toggle);
  return Val_unit;
}


/* open GL */

value sdlstub_GL_swap_buffers(value u) {
  SDL_GL_SwapBuffers();
  return Val_unit;
}


/* audio */

/*
value sdlstub_open_audio(value u) {
  SDL_OpenAudio();
  return Val_unit;
}
*/

value sdlstub_close_audio(value u) {
  SDL_CloseAudio();
  return Val_unit;
}

value sdlstub_unlock_audio(value u) {
  SDL_UnlockAudio();
  return Val_unit;
}

